@@ -1,13 +1,13 @@
 ;;; sideline.el --- Show information on the side  -*- lexical-binding: t; -*-
 
-;; Copyright (C) 2022-2024  Shen, Jen-Chieh
+;; Copyright (C) 2022-2026  Shen, Jen-Chieh
 ;; Created date 2022-06-13 22:08:26
 
 ;; Author: Shen, Jen-Chieh <jcs090218@gmail.com>
 ;; URL: https://github.com/emacs-sideline/sideline
-;; Package-Version: 0.2.0
-;; Package-Revision: 795673c66bb7
-;; Package-Requires: ((emacs "27.1") (ht "2.4"))
+;; Package-Version: 20260101.540
+;; Package-Revision: b4ada1ddf7da
+;; Package-Requires: ((emacs "28.1") (ht "2.4"))
 ;; Keywords: convenience
 
 ;; This file is NOT part of GNU Emacs.
@@ -45,6 +45,7 @@
 
 (require 'cl-lib)
 (require 'face-remap)
+(require 'mule-util)
 (require 'rect)
 (require 'subr-x)
 
@@ -58,12 +59,12 @@
 
 (defcustom sideline-backends-left nil
   "The list of active backends to display sideline on the left."
-  :type 'list
+  :type '(list symbol)
   :group 'sideline)
 
 (defcustom sideline-backends-right nil
   "The list of active backends to display sideline on the right."
-  :type 'list
+  :type '(list symbol)
   :group 'sideline)
 
 (defcustom sideline-order-left 'down
@@ -83,6 +84,11 @@
   :type 'boolean
   :group 'sideline)
 
+(defcustom sideline-truncate t
+  "Truncate sideline if the line width are wider than the window width."
+  :type 'boolean
+  :group 'sideline)
+
 (defface sideline-default
   '((((background light)) :foreground "DarkOrange")
     (t :foreground "yellow"))
@@ -143,6 +149,11 @@
   :type 'number
   :group 'sideline)
 
+(defcustom sideline-backend-delays nil
+  "Number of seconds to wait before showing sideline per backend."
+  :type '(list symbol)
+  :group 'sideline)
+
 (defcustom sideline-pre-render-hook nil
   "Hooks runs before rendering sidelines."
   :type 'hook
@@ -187,15 +198,51 @@
 (defvar-local sideline-render-this-command nil
   "If this is non-nil, re-render this command.")
 
+(defvar-local sideline--max-remain-spaces nil
+  "Hold the maximum remaining spaces when finding lines..")
+
+(defvar-local sideline--max-remain-spaces-line nil
+  "Hold the line information that has the maximum remaining spaces.")
+
+;;
+;; (@* "Render Flags" )
+;;
+
+(defvar-local sideline--ex-window nil
+  "Holds previous window.")
+
+(defvar-local sideline--ex-window-start nil
+  "Holds previous window start point; detect vertical scrolling.")
+
+(defvar-local sideline--ex-window-hscroll nil
+  "Holds previous window hscroll; detect horizontal scrolling.")
+
+(defvar-local sideline--ex-window-width nil
+  "Holds previous window width; detect size change.")
+
+(defvar-local sideline--ex-window-height nil
+  "Holds previous window height; detect size change.")
+
+(defvar-local sideline--ex-face-remapping-alist nil
+  "Holds previous face remapping alist.")
+
 ;;
 ;; (@* "Externals" )
 ;;
 
 (defvar overflow-newline-into-fringe)
 
-(declare-function string-pixel-width "subr-x.el")   ; TODO: remove this after 29.1
+(declare-function buffer-text-pixel-size "src/xdisp.c")
+
 (declare-function shr-string-pixel-width "shr.el")  ; TODO: remove this after 29.1
 
+(defun sideline--handle-externals ()
+  "Alter external packages to work with `sideline'."
+  (add-hook 'whitespace-mode-hook #'sideline-render-this)
+  ;; Tree-sitter
+  (add-hook 'ts-fold-on-fold-hook #'sideline-render-this)
+  (add-hook 'treesit-fold-on-fold-hook #'sideline-render-this))
+
 ;;
 ;; (@* "Entry" )
 ;;
@@ -207,12 +254,14 @@
   ;; XXX: Still don't know why local variable doesn't work!
   (progn
     (sideline--delete-ovs)
-    (setq-local sideline--overlays (make-hash-table)))
+    (setq-local sideline--overlays (make-hash-table)
+                sideline--backend-delay-timers (make-hash-table)))
   (setq sideline--render-data-wapp (make-hash-table)
         sideline--ex-bound-or-point t  ; reset, render immediately
         sideline--text-scale-mode-amount text-scale-mode-amount)
   (add-hook 'post-command-hook #'sideline--post-command -90 t)
   (add-hook 'before-revert-hook #'sideline--before-revert nil t)
+  (sideline--handle-externals)
   ;; Render immediately after reopened file!
   (sideline--post-command))
 
@@ -255,8 +304,8 @@
 (defmacro sideline--with-buffer-window (buffer-or-name &rest body)
   "Execute the forms in BODY with BUFFER-OR-NAME temporarily current."
   (declare (indent 1) (debug t))
-  `(when-let (((buffer-live-p ,buffer-or-name))
-              (window (get-buffer-window ,buffer-or-name)))
+  `(when-let* (((buffer-live-p ,buffer-or-name))
+               (window (get-buffer-window ,buffer-or-name)))
      (with-selected-window window
        (with-current-buffer ,buffer-or-name ,@body))))
 
@@ -357,7 +406,12 @@ Argument OFFSET is additional calculation from the right alignment."
               (if graphic-p
                   ;; If right fringe deactivated add 1 offset
                   (if (= 0 (nth 1 fringes)) 1 0)
-                1)))))))
+                1)
+              (if (or
+                   (bound-and-true-p whitespace-mode)
+                   (bound-and-true-p global-whitespace-mode))
+                  1
+                0)))))))
 
 (defun sideline--line-pixel-start ()
   "Return the pixel start of the line."
@@ -390,7 +444,12 @@ Argument OFFSET is additional calculation from the right alignment."
          (len (- end start)))
     (sideline--to-text-width len)))            ; to text space
 
-(defun sideline--calc-space (str-len on-left opposing-str-len)
+(defun sideline--exceed-display-p ()
+  "Return non-nil if exceed display."
+  (or sideline-force-display-if-exceeds
+      sideline-truncate))
+
+(defun sideline--calc-space (str-len on-left opposing-str-len bol)
   "Calculate space in current line.
 
 Argument STR-LEN is the string size.  Another argument OPPOSING-STR-LEN is the
@@ -405,22 +464,38 @@ calculate to the right side."
   (setq str-len (+ str-len opposing-str-len))
   ;; Start the calculation!
   (when-let* ((win-width (sideline--render-data :win-width))
-              ((or sideline-force-display-if-exceeds
-                   (<= str-len win-width)))
+              ((or (<= str-len win-width)
+                   (sideline--exceed-display-p)))
               (column-start (sideline--render-data :hscroll))
-              (pos-end (max (sideline--line-width) column-start)))
+              (pos-end (ignore-errors (sideline--line-width)))
+              (pos-end (max pos-end column-start)))
     (cond
      (on-left
-      (let ((pos-first (save-excursion (back-to-indentation) (current-column))))
-        (cond ((<= str-len (- pos-first column-start))
+      (let* ((pos-first (save-excursion (back-to-indentation) (current-column)))
+             ;; The variable `pos-first' cannot be lower than the `hscroll'.
+             (pos-first (max pos-first column-start))
+             (remain-spaces (- pos-first column-start)))
+        (cond ((<= str-len remain-spaces)
                (cons column-start pos-first))
               ((= pos-first pos-end)
-               (cons column-start win-width)))))
+               (cons column-start win-width))
+              ((and (sideline--exceed-display-p)
+                    (< sideline--max-remain-spaces remain-spaces))
+               (setq sideline--max-remain-spaces remain-spaces
+                     sideline--max-remain-spaces-line (cons bol
+                                                            (cons column-start pos-first)))
+               nil))))  ; Ensure return `nil'.
      (t
-      (let ((column-end (+ column-start win-width)))
-        (cond ((or sideline-force-display-if-exceeds
-                   (<= str-len (- column-end pos-end)))
-               (cons column-end pos-end))))))))
+      (let* ((column-end (+ column-start win-width))
+             (remain-spaces (- column-end pos-end)))
+        (cond ((<= str-len remain-spaces)
+               (cons column-end pos-end))
+              ((and (sideline--exceed-display-p)
+                    (< sideline--max-remain-spaces remain-spaces))
+               (setq sideline--max-remain-spaces remain-spaces
+                     sideline--max-remain-spaces-line (cons bol
+                                                            (cons column-end pos-end)))
+               nil)))))))
 
 (defun sideline--find-line (str-len on-left &optional direction exceeded)
   "Find a line where the string can be inserted.
@@ -444,21 +519,39 @@ available lines in both directions (up & down)."
       (while (not break-it)
         (if skip-first (setq skip-first nil)
           (forward-visible-line (if going-up -1 1)))
-        (unless (if going-up (<= (sideline--render-data :bol) (point))
-                  (<= (point) (sideline--render-data :eol)))
-          (setq break-it t))
-        (when-let* ((occ-bol (line-beginning-position))
-                    ((and (not (memq occ-bol occupied-lines))
-                          (not break-it)))
-                    (col (sideline--calc-space str-len on-left (sideline--opposing-str-len)))
-                    (pos-start (sideline--column-to-point (car col)))
-                    (pos-end   (sideline--column-to-point (cdr col)))
-                    ;; Skip virtual line from `truncate-lines'.
-                    ((= pos-start pos-end)))
+        ;; Reach window edge, means find nothing.
+        (unless break-it
+          (unless (if going-up
+                      (<= (sideline--render-data :win-start) (point))
+                    (<= (point) (sideline--render-data :win-end)))
+            (setq break-it t)))
+        ;; Reach buffer edge, means find nothing.
+        (unless break-it
+          (when (if going-up (bobp) (eobp))
+            (setq break-it t)))
+        (when-let*
+            ((occ-bol (line-beginning-position))
+             ((not (memq occ-bol occupied-lines)))
+             (col (or (sideline--calc-space str-len on-left
+                                            (sideline--opposing-str-len)
+                                            occ-bol)
+                      ;; Handle exceed display.
+                      (when (and exceeded  ; after search on both sides.
+                                 break-it  ; after reach edges.
+                                 sideline--max-remain-spaces-line)
+                        ;; Apply to BOL
+                        (setq occ-bol (car sideline--max-remain-spaces-line))
+                        (goto-char occ-bol)
+                        ;; Return column data.
+                        (cdr sideline--max-remain-spaces-line))))
+             (pos-start (sideline--column-to-point (car col)))
+             (pos-end   (sideline--column-to-point (cdr col)))
+             ((or on-left
+                  ;; Skip virtual line from `truncate-lines'.
+                  (= pos-start pos-end))))
           (setq data (list pos-start pos-end occ-bol))
           (setq break-it t)
-          (push occ-bol occupied-lines))
-        (when (if going-up (bobp) (eobp)) (setq break-it t))))
+          (push occ-bol occupied-lines))))
     (if on-left
         (setq sideline--occupied-lines-left occupied-lines)
       (setq sideline--occupied-lines-right occupied-lines))
@@ -573,25 +666,53 @@ FACE, NAME, ON-LEFT, and ORDER for details."
           (if on-left (format sideline-format-left text)
             (format sideline-format-right text))))
        (len-title (sideline--str-len title))
-       (data (sideline--find-line len-title on-left order))
+       (data (let ((sideline--max-remain-spaces 0)          ; Reset before use.
+                   (sideline--max-remain-spaces-line nil))  ; Reset before use.
+               (sideline--find-line len-title on-left order)))
        (pos-start (nth 0 data)) (pos-end (nth 1 data)) (occ-pt (nth 2 data))
-       (offset (- 0 (sideline--render-data :hscroll)))
+       (hscroll (sideline--render-data :hscroll))
+       (offset (- 0 hscroll))
+       ;; Truncate
+       (title
+        (or (and sideline-truncate
+                 (let* ((win-width (sideline--render-data :win-width))
+                        (end-col (save-excursion
+                                   (goto-char pos-end)
+                                   (current-column)))
+                        (available-space (if on-left
+                                             (if (zerop end-col)
+                                                 win-width
+                                               (- end-col hscroll 1))
+                                           (- win-width (- end-col hscroll))))
+                        (suffix (copy-sequence (truncate-string-ellipsis))))
+                   (when (< available-space len-title)
+                     (set-text-properties 0 (length suffix)
+                                          (text-properties-at (1- (length title)) title)
+                                          suffix)
+                     (truncate-string-to-width title available-space 0 nil
+                                               suffix))))
+            title))
+       ;; Align left/right
        (str (concat
              (unless on-left
                (propertize " "
-                           'display `((space :align-to
-                                             (- right ,(sideline--align-right title offset)))
-                                      (space :width 0))
+                           'display
+                           `((space :align-to (- right ,(sideline--align-right title offset)))
+                             (space :width 0))
                            `cursor t))
              title)))
+
     ;; Create overlay
     (let* ((len-str (length str))
-           (empty-ln (= pos-start pos-end))
-           (ov (make-overlay pos-start (if empty-ln pos-start (+ pos-start len-str))
+           (left-empty-ln (= pos-start pos-end))
+           (ov (make-overlay pos-start
+                             (if left-empty-ln pos-start
+                               ;; Cannot exceed `pos-end'.
+                               (min pos-end (+ pos-start len-str)))
                              nil t t)))
       (cond (on-left
-             (if empty-ln
-                 (overlay-put ov 'before-string str)
+             (if left-empty-ln
+                 (overlay-put ov 'before-string (concat (spaces-string hscroll) str))
                (overlay-put ov 'display str)
                (overlay-put ov 'invisible t)))
             (t (overlay-put ov 'before-string str)))
@@ -639,29 +760,60 @@ Argument ORDER determined the search order for going up or down."
 ;; (@* "Core" )
 ;;
 
+(defvar-local sideline--backend-delay-timers nil
+  "Delay timer per backend.")
+
 (defun sideline--call-backend (backend command)
   "Return BACKEND's result with COMMAND."
   (funcall backend command))
 
+(defun sideline--render-backend (backend on-left order)
+  "Render for single BACKEND.
+
+All arguments BACKEND, ON-LEFT and ORDER are extracted from the function
+`sideline--render-backends'."
+  (when-let* ((candidates (sideline--call-backend backend 'candidates))
+              (buffer (current-buffer)))  ; for async check
+    (if (eq (car candidates) :async)
+        (funcall (cdr candidates)
+                 (lambda (cands &rest _)
+                   (sideline--with-buffer-window buffer
+                     (when sideline-mode
+                       ;; XXX: The bug occurs when you have two or more indentical
+                       ;; windows (same buffer and same cursor position) and
+                       ;; frequently change between different buffers.
+                       ;;
+                       ;; To solve this, just record the "real" rendering
+                       ;; buffer in async. So when you switch back to the same
+                       ;; buffer, the sideline can adjust to find the correct
+                       ;; buffer in the function `sideline--do-render-p' since
+                       ;; it has been flag here.
+                       (setq sideline--ex-window (selected-window))
+                       (sideline--render-candidates cands backend on-left order)))))
+      (sideline--render-candidates candidates backend on-left order))))
+
 (defun sideline--render-backends (backends on-left)
   "Render a list of BACKENDS.
 
-If argument ON-LEFT is non-nil, it will align to the left instead of right."
+If argument ON-LEFT is non-nil, it will align to the left
+instead of right."
   (dolist (data backends)
-    (let* ((is-cons (consp data))
-           (backend (if is-cons (car data) data))
-           (order (if is-cons (cdr data)  ; configured
-                    ;; fallback to default
-                    (if on-left sideline-order-left sideline-order-right)))
-           (candidates (sideline--call-backend backend 'candidates))
-           (buffer (current-buffer)))  ; for async check
-      (if (eq (car candidates) :async)
-          (funcall (cdr candidates)
-                   (lambda (cands &rest _)
-                     (sideline--with-buffer-window buffer
-                       (when sideline-mode
-                         (sideline--render-candidates cands backend on-left order)))))
-        (sideline--render-candidates candidates backend on-left order)))))
+    (let ((is-cons (consp data)))
+      (when-let* ((backend (if is-cons (car data) data))
+                  (backend (if (symbolp backend) backend
+                             (eval backend)))
+                  (order (if is-cons (cdr data)  ; configured
+                           ;; fallback to default
+                           (if on-left sideline-order-left sideline-order-right))))
+        ;; Handle backend delay here.
+        (if-let* ((delay (alist-get backend sideline-backend-delays)))
+            (progn
+              (sideline--kill-timer (ht-get sideline--backend-delay-timers backend))
+              (ht-set sideline--backend-delay-timers
+                      backend
+                      (run-with-idle-timer delay nil #'sideline--render-backend
+                                           backend on-left order)))
+          (sideline--render-backend backend on-left order))))))
 
 (defun sideline-stop-p ()
   "Return non-nil if the sideline should not be display."
@@ -673,64 +825,63 @@ If argument ON-LEFT is non-nil, it will align to the left instead of right."
   "Get render data by PROP."
   (plist-get sideline--render-data prop))
 
+(defun sideline--render-data-update (&rest _)
+  "Update the render data once."
+  (setq sideline--render-data
+        `( :win-end   ,(sideline--window-end)
+           :win-start ,(window-start)
+           :hscroll   ,(sideline--window-hscroll)
+           :win-width ,(sideline--window-width))))
+
 (defun sideline-render (&optional buffer)
   "Render sideline once in the BUFFER of WINDOW."
   (sideline--with-buffer-window (or buffer (current-buffer))
     (unless (funcall sideline-inhibit-display-function)
-      (setq sideline--render-data
-            `( :eol ,(sideline--window-end)
-               :bol ,(window-start)
-               :hscroll ,(sideline--window-hscroll)
-               :win-width ,(sideline--window-width)))
+      (sideline--render-data-update)
       (run-hooks 'sideline-pre-render-hook)
       (sideline--render-backends sideline-backends-left t)
       (sideline--render-backends sideline-backends-right nil)
       (run-hooks 'sideline-post-render-hook))))
 
-(defvar-local sideline--delay-timer nil
-  "Timer for delay.")
-
-(defvar-local sideline--ex-window nil
-  "Holds previous window.")
-
-(defvar-local sideline--ex-window-start nil
-  "Holds previous window start point; this will detect vertical scrolling.")
-
-(defvar-local sideline--ex-window-hscroll nil
-  "Holds previous window hscroll; this will detect horizontal scrolling.")
-
-(defvar-local sideline--ex-face-remapping-alist nil
-  "Holds previous face remapping alist.")
-
 (defun sideline--do-render-p ()
   "Return non-nil if we should re-render sidelines in the post-command."
-  (let ((bound-or-point (or (bounds-of-thing-at-point 'symbol) (point)))
-        (window (selected-window))
-        (win-start (window-start))
-        (win-hscroll (window-hscroll))
+  (let ((bound-or-point  (or (bounds-of-thing-at-point 'symbol)
+                             (point)))
+        (window          (selected-window))
+        (win-start       (window-start))
+        (win-hscroll     (window-hscroll))
+        (win-width       (window-width))
+        (win-height      (window-height))
         (remapping-alist face-remapping-alist))
     (when  ; conditions allow to re-render sidelines
-        (or (not (equal sideline--ex-bound-or-point bound-or-point))
-            (not (equal sideline--text-scale-mode-amount text-scale-mode-amount))
-            (not (equal sideline--ex-window window))
-            (not (equal sideline--ex-window-start win-start))
-            (not (equal sideline--ex-window-hscroll win-hscroll))
+        (or (not (equal sideline--ex-bound-or-point       bound-or-point))
+            (not (equal sideline--text-scale-mode-amount  text-scale-mode-amount))
+            (not (equal sideline--ex-window               window))
+            (not (equal sideline--ex-window-start         win-start))
+            (not (equal sideline--ex-window-hscroll       win-hscroll))
+            (not (equal sideline--ex-window-width         win-width))
+            (not (equal sideline--ex-window-height        win-height))
             (not (equal sideline--ex-face-remapping-alist remapping-alist))
             sideline-render-this-command)
       ;; update
-      (setq sideline--ex-bound-or-point bound-or-point
-            sideline--text-scale-mode-amount text-scale-mode-amount
-            sideline--ex-window window
-            sideline--ex-window-start win-start
-            sideline--ex-window-hscroll win-hscroll
+      (setq sideline--ex-bound-or-point       bound-or-point
+            sideline--text-scale-mode-amount  text-scale-mode-amount
+            sideline--ex-window               window
+            sideline--ex-window-start         win-start
+            sideline--ex-window-hscroll       win-hscroll
+            sideline--ex-window-width         win-width
+            sideline--ex-window-height        win-height
             sideline--ex-face-remapping-alist remapping-alist
-            sideline-render-this-command nil)
+            sideline-render-this-command      nil)
       t)))
 
 (defun sideline--before-revert (&rest _)
   "Before revert."
   (sideline--reset))
 
+(defvar-local sideline--delay-timer nil
+  "Delay timer to render sideline.")
+
 (defun sideline--post-command ()
   "Post command."
   (when (sideline--do-render-p)
@@ -739,6 +890,9 @@ If argument ON-LEFT is non-nil, it will align to the left instead of right."
     (setq sideline--delay-timer
           (run-with-idle-timer sideline-delay nil #'sideline-render
                                (current-buffer)))
+    ;; XXX: Reset hook can potentially re-render sideline; update the render
+    ;; data once to ensure display's correctioness.
+    (sideline--render-data-update)
     (run-hooks 'sideline-reset-hook)))
 
 ;;;###autoload
